package pwc.taxtech.atms.vat.service.impl;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;
import pwc.taxtech.atms.common.datasource.ShardingContextHolder;
import pwc.taxtech.atms.constant.enums.EnumServiceType;
import pwc.taxtech.atms.dto.OperationResultDto;
import pwc.taxtech.atms.dto.vatdto.*;
import pwc.taxtech.atms.entitiy.*;
import pwc.taxtech.atms.exception.NotSupportedException;
import pwc.taxtech.atms.service.impl.CellConfigTranslater;
import pwc.taxtech.atms.vat.entity.*;
import pwc.taxtech.atms.vat.service.ReportService;

import java.util.*;
import java.util.stream.Collectors;

@Service
public class ReportServiceImpl extends VatAbstractService implements ReportService {
    @Override
    public OperationResultDto<List<ReportDto>> getReportTemplate(String projectID, EnumServiceType serviceType, Integer periodParam) {
        int period = periodParam != null ? periodParam : 0;
        OperationResultDto<List<ReportDto>> operationResult = new OperationResultDto<>();
        Project project = projectMapper.selectByPrimaryKey(projectID);
        String organizationID = project == null ? null : project.getOrganizationID();
        if (organizationID == null) {
            operationResult.setResult(false);
            operationResult.setResultMsg("NoOrganization");
            return operationResult;
        }

        Long templateGroupID;

        /*
        对于VAT，从Rule Engine中获取TemplateGroupID
        对于其他Service Type，直接从ProjectServiceType中获取
         */
        if (serviceType.equals(EnumServiceType.VAT)) {
            TaxPayerReportRuleExample example = new TaxPayerReportRuleExample();
            example.createCriteria().andOrganizationIdLike(organizationID);
            List<TaxPayerReportRule> taxPayerReportRules = taxPayerReportRuleMapper.selectByExample(example);
            Optional<TaxPayerReportRule> taxPayerReportRule = taxPayerReportRules.stream().findFirst();
            if (!taxPayerReportRule.isPresent()) {
                Optional<TaxPayerReportRule> defaultRule = taxPayerReportRules.stream().filter(x -> x.getIsDefault()).findFirst();
                if (!defaultRule.isPresent()) {
                    operationResult.setResultMsg("NoDefaultRule");
                    return operationResult;
                }

                templateGroupID = defaultRule.get().getTemplateGroupId();
            } else {
                templateGroupID = taxPayerReportRule.get().getTemplateGroupId();
            }

        } else {
            String serviceTypeStr = serviceType.getCode().toString();
            ProjectServiceTypeExample projectServiceTypeExample = new ProjectServiceTypeExample();
            projectServiceTypeExample.createCriteria().andServiceTypeIDEqualTo(serviceTypeStr).andProjectIDEqualTo(projectID);
            Optional<Long> tempVlaue = projectServiceTypeMapper.selectByExample(projectServiceTypeExample).stream().map(ProjectServiceType::getTemplateGroupID).findFirst();
            if (tempVlaue.isPresent()) {
                templateGroupID = tempVlaue.get();
            } else {
                templateGroupID = 0L;
            }
        }

        String dbName = ShardingContextHolder.getDataSourceKey();
        Map<String, Object> map = new HashMap<>();
        map.put("dbName", dbName);
        map.put("period", period);
        map.put("projectID", projectID);
        map.put("templateGroupID", templateGroupID);

        List<ReportDto> reportDtos = templateMapper.getTemplateLeftJoinReport(map);
        if (reportDtos.isEmpty()) {
            operationResult.setResultMsg("No Template");
            return operationResult;
        }

        operationResult.setResult(true);
        operationResult.setData(reportDtos);
        return operationResult;
    }

    @Override
    public OperationResultDto updateConfig(String projectId, Integer period, Boolean ifDeleteManualDataSource, String generator) {
        OperationResultDto result = new OperationResultDto();

        if (period == null) {
            result.setResultMsg("peirod is null");
            return result;
        }

        Project project = projectMapper.selectByPrimaryKey(projectId);

        if (project == null) {
            result.setResultMsg("NoProject");
            return result;
        }

        List<Long> exceptTemplateIDs = templateMapper.getIDsForExceptTemplate();

        String strExceptTemplateIDs = StringUtils.join(exceptTemplateIDs, ",");

        //todo: according to projectId get the templateGroupId
        Long templateGroupID = projectMapper.getTemplateGroupIdByProject(projectId, EnumServiceType.VAT.getCode());
        if (templateGroupID != null && templateGroupID != 0) {
            // 先进行数据清理，包括period开头的所有报表配置表 条件Period
            dataSourceMapper.clearFormulaBlockWithPeriod(period);
            dataSourceMapper.clearTaxRuleSettingWithPeriod(period);
            dataSourceMapper.clearCellTemplateWithPeriod(period, strExceptTemplateIDs);
            dataSourceMapper.clearCellTemplateConfigWithPeriod(period, strExceptTemplateIDs);
            dataSourceMapper.clearTemplateWithPeriod(period, strExceptTemplateIDs);
            dataSourceMapper.clearTaxPayerReportWithPeriod(period);
            dataSourceMapper.clearDataSourceWithPeriod(period);
            dataSourceMapper.clearDataSourceDetailWithPeriod(period);
            dataSourceMapper.clearCellDataWithPeriod(strExceptTemplateIDs);

            // 根据templategroupid 把 template 插入到 periodTemplate
            TemplateExample example = new TemplateExample();
            example.createCriteria().andTemplateGroupIdEqualTo(templateGroupID);
            List<Template> templates = templateMapper.selectByExample(example);
            List<PeriodTemplate> periodTemplateList = new ArrayList<>();
            for (Template template : templates) {
                PeriodTemplate periodTemplate = new PeriodTemplate();
                periodTemplate.setId(distributedIDService.nextId());
                periodTemplate.setPeriod(period);
                periodTemplate.setName(template.getName());
                periodTemplate.setCode(template.getCode());
                periodTemplate.setPath(template.getPath());
                periodTemplate.setReportType(template.getReportType());
                periodTemplate.setTemplateGroupId(template.getTemplateGroupId());
                periodTemplate.setOrderIndex(template.getOrderIndex());
                periodTemplate.setCreateTime(template.getCreateTime());
                periodTemplate.setUpdateTime(template.getUpdateTime());
                periodTemplate.setIsSystemType(template.getIsSystemType());
                periodTemplate.setIsActiveAssociation(template.getIsActiveAssociation());
                periodTemplate.setParentId(template.getParentId());
                periodTemplate.setTemplateId(template.getId());
                periodTemplate.setCreateBy(template.getCreateBy());
                periodTemplate.setUpdateBy(template.getUpdateBy());
                periodTemplateList.add(periodTemplate);
            }

            List<Long> templateIDList = periodTemplateList.stream().map(PeriodTemplate::getTemplateId).collect(Collectors.toList());
            CellTemplateExample cellTemplateExample = new CellTemplateExample();
            cellTemplateExample.createCriteria().andReportTemplateIdIn(templateIDList);
            List<CellTemplate> cellTemplateList = cellTemplateMapper.selectByExample(cellTemplateExample);
            List<PeriodCellTemplate> periodCellTemplateList = new ArrayList<>();
            for (CellTemplate cellTemplate : cellTemplateList) {
                PeriodCellTemplate periodCellTemplate = new PeriodCellTemplate();
                periodCellTemplate.setId(distributedIDService.nextId());
                periodCellTemplate.setPeriod(period);
                periodCellTemplate.setReportTemplateId(cellTemplate.getReportTemplateId());
                periodCellTemplate.setRowIndex(cellTemplate.getRowIndex());
                periodCellTemplate.setRowName(cellTemplate.getRowName());
                periodCellTemplate.setColumnIndex(cellTemplate.getColumnIndex());
                periodCellTemplate.setColumnName(cellTemplate.getColumnName());
                periodCellTemplate.setComment(cellTemplate.getComment());
                periodCellTemplate.setCreateTime(cellTemplate.getCreateTime());
                periodCellTemplate.setUpdateTime(cellTemplate.getUpdateTime());
                periodCellTemplate.setCellTemplateId(cellTemplate.getId());
                periodCellTemplate.setDataType(cellTemplate.getDataType());
                periodCellTemplate.setIsReadOnly(cellTemplate.getIsReadOnly());
                periodCellTemplate.setCopyFromId(cellTemplate.getCopyFromId());
                periodCellTemplate.setCreateBy(cellTemplate.getCreateBy());
                periodCellTemplate.setUpdateBy(cellTemplate.getUpdateBy());
                periodCellTemplateList.add(periodCellTemplate);
            }

            List<Long> cellTemplateIDList = periodCellTemplateList.stream().map(PeriodCellTemplate::getCellTemplateId).collect(Collectors.toList());
            CellTemplateConfigExample cellTemplateConfigExample = new CellTemplateConfigExample();
            cellTemplateConfigExample.createCriteria().andReportTemplateIdIn(templateIDList).andCellTemplateIdIn(cellTemplateIDList);
            List<CellTemplateConfig> cellTemplateConfigList = cellTemplateConfigMapper.selectByExample(cellTemplateConfigExample);
            List<PeriodCellTemplateConfig> periodCellTemplateConfigList = new ArrayList<>();
            for (CellTemplateConfig cellTemplateConfig : cellTemplateConfigList) {
                PeriodCellTemplateConfig periodCellTemplateConfig = new PeriodCellTemplateConfig();
                periodCellTemplateConfig.setId(distributedIDService.nextId());
                periodCellTemplateConfig.setPeriod(period);
                periodCellTemplateConfig.setCellTemplateId(cellTemplateConfig.getCellTemplateId());
                periodCellTemplateConfig.setReportTemplateId(cellTemplateConfig.getReportTemplateId());
                periodCellTemplateConfig.setDataSourceType(cellTemplateConfig.getDataSourceType());
                periodCellTemplateConfig.setFormula(cellTemplateConfig.getFormula());
                periodCellTemplateConfig.setParsedFormula(StringUtils.EMPTY);
                periodCellTemplateConfig.setFormulaDescription(cellTemplateConfig.getFormulaDescription());
                periodCellTemplateConfig.setAccountCodes(cellTemplateConfig.getAccountCodes());
                periodCellTemplateConfig.setInvoiceType(cellTemplateConfig.getInvoiceType());
                periodCellTemplateConfig.setTaxRate(cellTemplateConfig.getTaxRate());
                periodCellTemplateConfig.setInvoiceAmountType(cellTemplateConfig.getInvoiceAmountType());
                periodCellTemplateConfig.setModelIds(cellTemplateConfig.getModelIds());
                periodCellTemplateConfig.setCreateBy(cellTemplateConfig.getCreateBy());
                periodCellTemplateConfig.setCreateTime(cellTemplateConfig.getCreateTime());
                periodCellTemplateConfig.setUpdateBy(cellTemplateConfig.getUpdateBy());
                periodCellTemplateConfig.setUpdateTime(cellTemplateConfig.getUpdateTime());
                periodCellTemplateConfig.setInvoiceCategory(cellTemplateConfig.getInvoiceCategory());
                periodCellTemplateConfig.setFormulaDataSource(cellTemplateConfig.getFormulaDataSource());
                periodCellTemplateConfig.setValidation(cellTemplateConfig.getValidation());
                periodCellTemplateConfig.setParsedValidation(StringUtils.EMPTY);
                periodCellTemplateConfig.setValidationDescription(cellTemplateConfig.getValidationDescription());
                periodCellTemplateConfig.setVoucherKeyword(cellTemplateConfig.getVoucherKeyword());
                periodCellTemplateConfig.setCellTemplateConfigId(cellTemplateConfig.getId());
                periodCellTemplateConfigList.add(periodCellTemplateConfig);
            }

            if (periodCellTemplateConfigList != null) {
                periodCellTemplateConfigList.forEach(a -> {
                    if (StringUtils.isNotBlank(a.getAccountCodes())) {
                        String[] acctCodes = ",".split(a.getAccountCodes());
                        EnterpriseAccountExample enterpriseAccountExample = new EnterpriseAccountExample();
                        enterpriseAccountExample.createCriteria().andStdCodeIn(Arrays.asList(acctCodes));
                        List<EnterpriseAccount> enterpriseAccounts = enterpriseAccountMapper.selectByExample(enterpriseAccountExample);
                        if (!enterpriseAccounts.isEmpty()) {
                            List<String> acctCodeList = enterpriseAccounts.stream().map(EnterpriseAccount::getCode).collect(Collectors.toList());
                            StringBuilder codes = new StringBuilder();
                            for (String code : acctCodeList) {
                                codes.append(code).append(",");
                            }
                            a.setAccountCodes(StringUtils.removeEnd(codes.toString(), ","));
                        }
                    }
                });
            }

            //todo: save config data to DB
            periodTemplateMapper.batchInsert(periodTemplateList);
            periodCellTemplateMapper.batchInsert(periodCellTemplateList);
            periodCellTemplateConfigMapper.batchInsert(periodCellTemplateConfigList);
        } else {
            result.setResult(true);
            result.setResultMsg("there is no templateGroup");
            return result;
        }
        result.setResult(true);
        return result;
    }

    @Override
    public OperationResultDto generateData(String projectId, EnumServiceType serviceType, Boolean ifDeleteManualDataSource, Integer periodParam, Integer reportType, String generator) {
        OperationResultDto operationResultDto = new OperationResultDto();

        if (serviceType.equals(EnumServiceType.VAT) && (periodParam == null && periodParam <= 0)) {
            operationResultDto.setResultMsg("PeriodRequiredForVAT");
            return operationResultDto;
        }

        if (serviceType != EnumServiceType.VAT) {
            // 暂不支持非CIT/VAT service批量后端生成报表
            throw new NotSupportedException();
        }

        int period = periodParam != null ? periodParam : 0;
        String serviceTypeStr = serviceType.getCode().toString();

        ProjectServiceTypeExample projectServiceTypeExample = new ProjectServiceTypeExample();
        projectServiceTypeExample.createCriteria().andServiceTypeIDEqualTo(serviceTypeStr).andProjectIDEqualTo(projectId);
        Optional<Long> templateGroupID = projectServiceTypeMapper.selectByExample(projectServiceTypeExample).stream().map(ProjectServiceType::getTemplateGroupID).findFirst();
        if (templateGroupID == null) {
            operationResultDto.setResultMsg("TemplateGroupNotExist");
            return operationResultDto;
        }

        List<Long> templateIds;
        if (reportType != null) {
            PeriodTemplateExample periodTemplateExample = new PeriodTemplateExample();
            periodTemplateExample.createCriteria().andTemplateGroupIdEqualTo(templateGroupID.get()).andPeriodEqualTo(period).andIsActiveAssociationEqualTo(true).andReportTypeEqualTo(reportType);
            templateIds = periodTemplateMapper.selectByExample(periodTemplateExample).stream().map(PeriodTemplate::getId).collect(Collectors.toList());
        } else {
            PeriodTemplateExample periodTemplateExample = new PeriodTemplateExample();
            periodTemplateExample.createCriteria().andTemplateGroupIdEqualTo(templateGroupID.get()).andPeriodEqualTo(period).andIsActiveAssociationEqualTo(true);
            templateIds = periodTemplateMapper.selectByExample(periodTemplateExample).stream().map(PeriodTemplate::getId).collect(Collectors.toList());
        }

        String rslt = reportGenerator.generateData(projectId, templateIds, ifDeleteManualDataSource, null, periodParam, generator);
        if (StringUtils.isBlank(rslt)) {
            operationResultDto.setResultMsg("ReportGenerateFailed!");
            return operationResultDto;
        }
        operationResultDto.setResult(true);
        return operationResultDto;
    }

    @Override
    public List<CellTemplateReferenceDto> getTemplateReferences(int period) {
        return new ArrayList<>();
    }

    @Override
    public OperationResultDto<ReportDataDto> getCellData(Long reportId) {
        OperationResultDto resultDto = new OperationResultDto();
        ReportDataDto dataDto = new ReportDataDto();

        Report report = reportMapper.selectByPrimaryKey(reportId);
        if (report == null) {
            resultDto.setResult(false);
            resultDto.setResultMsg("NoReport");
            return resultDto;
        }

        PeriodTemplateExample example = new PeriodTemplateExample();
        example.createCriteria().andTemplateIdEqualTo(report.getTemplateId()).andPeriodEqualTo(report.getPeriod());
        Optional<PeriodTemplate> reportTemplate = periodTemplateMapper.selectByExample(example).stream().findFirst();
        if (!reportTemplate.isPresent()) {
            resultDto.setResult(false);
            resultDto.setResultMsg("NoReportTemplate");
            return resultDto;
        }

        Project project = projectMapper.selectByPrimaryKey(report.getProjectId());
        if (project == null) {
            resultDto.setResult(false);
            resultDto.setResultMsg("NoProject");
            return resultDto;
        }

        PeriodTemplateExample periodTemplateExample = new PeriodTemplateExample();
        periodTemplateExample.createCriteria()
                .andTemplateGroupIdEqualTo(reportTemplate.get().getTemplateGroupId())
                .andPeriodEqualTo(report.getPeriod())
                .andIsActiveAssociationEqualTo(true);
        List<PeriodTemplate> reportTemplates = periodTemplateMapper.selectByExample(example);


        List<Long> templateIDs = new ArrayList<>();
        templateIDs.add(report.getTemplateId());
        List<CellCalcInfoDto> cellCfgList = reportGenerator.getCellCalcInfo(templateIDs, report.getPeriod());

        CellDataExample cellDataExample = new CellDataExample();
        cellDataExample.createCriteria().andReportIdEqualTo(reportId);

        List<CellData> currentCellDataList = cellDataMapper.selectByExample(cellDataExample);

        PeriodFormulaBlockExample periodFormulaBlockExample = new PeriodFormulaBlockExample();
        periodFormulaBlockExample.createCriteria().andPeriodEqualTo(report.getPeriod());
        List<PeriodFormulaBlock> formulaBlocks = periodFormulaBlockMapper.selectByExample(periodFormulaBlockExample);
        dataDto.setFormulaBlocks(formulaBlocks);
        //todo: 关键字数据源稍后再加
        // 获取通过KeyValueDataID关联的数据源，即公式数据源
        //var formulaDataSourceList = (from ds in _dbprojectContext.Set<DataSource>()
        //join cd in _dbprojectContext.Set<CellData>()
        //on ds.KeyValueDataID equals cd.ID
        //where cd.ReportID == reportID && ds.Type == (int)CellDataSourceType.Formula
        //select new
        //{
        //    DataSource = ds,
        //            CellTemplateID = cd.CellTemplateID,
        //            CellDataID = cd.ID
        //}).AsNoTracking().ToList();

        //todo: 模型数据源稍后再加
        // 获取通过ModelBackFillAmount关联的数据源，即模型数据源
        // TODO: Check模型回填数据源是否会有多对多的情况出现，如有需distinct
        //var modelDataSourceList = (from ds in _dbprojectContext.Set<DataSource>()
        //join fa in _dbprojectContext.Set<ModelBackFillAmount>()
        //on ds.ID equals fa.DataSourceID
        //join cd in _dbprojectContext.Set<CellData>()
        //on fa.CellDataID equals cd.ID
        //where cd.ReportID == reportID && ds.Type == (int)FormulaDataSourceType.ModelDatasource
        //select new
        //{
        //    DataSource = ds,
        //            OperationType = (int)DataSourceOperationType.Plus,
        //            CellTemplateID = cd.CellTemplateID,
        //            CellDataID = cd.ID
        //}).AsNoTracking().ToList();

        //公式计算数据源 20180711 完成
        List<DataSourceExtendDto> manualDataSourceList = dataSourceMapper.getFormulaDataSource(reportId);
        Map<Long, DataSourceDto> datasource = new IdentityHashMap<>();
        manualDataSourceList.forEach(a -> {
            DataSourceDto dataSourceDto = new DataSourceDto();
            dataSourceDto.setId(a.getId());
            dataSourceDto.setAmount(a.getAmount());
            dataSourceDto.setName(a.getName());
            dataSourceDto.setOperationType(a.getOperationType());
            dataSourceDto.setDataSourceType(a.getType());
            dataSourceDto.setDescription(a.getDescription());
            dataSourceDto.setCreateTime(a.getCreateTime());
            datasource.put(a.getCellTemplateID(), dataSourceDto);
        });
        dataDto.setManualDataSources(datasource);

        List<CellDataDto> cellDataDtoList = new ArrayList<>();
        cellCfgList.forEach(a -> {
            CellDataDto cellDataDto = new CellDataDto();
            cellDataDto.setColumnIndex(a.getColumnIndex());
            cellDataDto.setColumnName(a.getColumnName());
            cellDataDto.setRowIndex(a.getRowIndex());
            cellDataDto.setRowName(a.getRowName());
            cellDataDto.setCellTemplateID(a.getCellTemplateID());
            cellDataDto.setReportTemplateID(a.getReportTemplateID());
            cellDataDto.setDataType(a.getDataType());
            cellDataDto.setReportID(reportId);
            cellDataDto.setIsReadOnly(a.getIsReadOnly());
            Optional<CellData> celldata = currentCellDataList.stream()
                    .filter(x -> x.getCellTemplateId().equals(a.getCellTemplateID())).findFirst();
            celldata.ifPresent(cellData -> cellDataDto.setCellID(cellData.getId()));
            cellDataDto.setFormula(a.getFormula());
            celldata.ifPresent(cellData -> cellDataDto.setCellValue(cellData.getData()));
            celldata.ifPresent(cellData -> cellDataDto.setFormulaExp(cellData.getFormulaExp()));
            cellDataDto.setCellTemplateConfig(CellConfigTranslater.getConfigDto(a));
            List<Map.Entry<Long, DataSourceDto>> entryList = datasource.entrySet().stream()
                    .filter(y -> y.getKey().equals(a.getCellTemplateID())).collect(Collectors.toList());
            List<DataSourceDto> dataSourceDtoList = new ArrayList<>();
            entryList.forEach(z -> dataSourceDtoList.add(z.getValue()));
            cellDataDto.setDataSourceList(dataSourceDtoList);
            cellDataDtoList.add(cellDataDto);
        });

        ModifiedReportCellExample modifiedReportCellExample = new ModifiedReportCellExample();
        modifiedReportCellExample.createCriteria().andReportIdEqualTo(reportId);
        List<ModifiedReportCell> modifiedReportCellList = modifiedReportCellMapper.selectByExample(modifiedReportCellExample);
        modifiedReportCellList.forEach(mrc -> {
            Optional<CellDataDto> cdd = cellDataDtoList.stream()
                    .filter(cellDataDto -> cellDataDto.getReportID().equals(mrc.getReportId())
                            && cellDataDto.getRowIndex() == mrc.getRow()
                            && cellDataDto.getColumnIndex() == mrc.getCol()).findFirst();
            if (cdd.isPresent()) {
                if (!mrc.getOriginalValue().equals(mrc.getValue())) {
                    cdd.get().setModifiedReportCell(mrc);
                    cdd.get().setCellValue(mrc.getValue());
                    cdd.get().setIsModified(true);
                }
            }
        });
        dataDto.setCellData(cellDataDtoList);
        resultDto.setData(dataDto);
        resultDto.setResult(true);
        return resultDto;
    }

    @Override
    public OperationResultDto<Report> getReportByTemplate(Long templateId, Integer period) {
        OperationResultDto resultDto = new OperationResultDto();
        ReportExample example = new ReportExample();
        example.createCriteria().andTemplateIdEqualTo(templateId).andPeriodEqualTo(period);
        Optional<Report> report = reportMapper.selectByExample(example).stream().findFirst();
        if (report.isPresent()) {
            resultDto.setResult(true);
            resultDto.setData(report.get());
        } else {
            resultDto.setResult(false);
            resultDto.setResultMsg("can not find report");
        }
        return resultDto;
    }
}
